# RefreshLess implementation using the [Hotwire Turbo library](https://turbo.hotwired.dev/)

## Requirements

* [Drupal core](https://www.drupal.org/project/drupal) 10.2 or newer

* [PHP](https://www.php.net/) 8.1 or newer

* [Composer](https://getcomposer.org/)

* [`cweagans/composer-patches` 1.x](https://github.com/cweagans/composer-patches/tree/1.x)

We generally recommend [DDEV](https://ddev.com/) because it provides most or all of the above for you out of the box, and [helps you get a Drupal project up and running](https://www.drupal.org/docs/develop/local-server-setup/docker-based-development-environments-for-macos-linux-and-windows) much faster while also isolating it from the rest of your operating system via containerization.

## Installation

Before attempting to install this, we strongly recommend enabling dependencies
to patch Drupal core; see [Known issues](#known-issues) for what we patch. Run
the following to set that up:

```shell
composer require 'cweagans/composer-patches:^1.7'
composer config extra.enable-patching true
composer config --json --merge extra.patchLevel '{"drupal/core": "-p2"}'
```

Then just `composer require` this module and core should be patched; if patches were not attempted, you can often try running `composer install` to give it a nudge to do so.

Finally, install this module in Drupal and it should take over handling page
requests.

## Known issues

* JavaScript aggregation works but there are limitations and steps required to enable it:

  * We force all JavaScript into the `<head>` but add the `defer` attribute as [recommended by the Turbo documentation](https://turbo.hotwired.dev/handbook/building#loading-your-application%E2%80%99s-javascript-bundle); this is necessary to prevent Turbo re-evaluating `<script>` elements that it finds in the `<body>`; [core is patched to allow aggregating JavaScript with the `defer` attribute](https://www.drupal.org/project/drupal/issues/1587536).

  * [We alter the core JavaScript aggregation behaviour to be additive during Turbo navigation](https://www.drupal.org/project/refreshless/issues/3414538), tracking what libraries are already loaded on a page in a similar way to Drupal's Ajax system. Because this can have multiple variations per page, Drupal's anonymous page caching is disabled for Turbo requests, though the dynamic page cache and render cache are unaffected.

  * This module's JavaScript must be grouped into a separate aggregate group to prevent it potentially being evaluated more than once if our additive aggregation doesn't behave as expected; [core is patched to allow explicitly defining an aggregation group](https://www.drupal.org/project/drupal/issues/3232810).

  * Turbo itself is not aggregated/minified at the moment as it starts to behave unpredictably and even attach more than once to a page if aggregated.

## Development

### Requirements

In addition to the main requirements above, the following is required to update
Turbo and patch it:

* A front-end package manager that supports both [the Plug'n'Play install strategy](https://yarnpkg.com/features/pnp) and [workspaces](https://yarnpkg.com/features/workspaces):

  * [Yarn 3 or newer](https://yarnpkg.com/getting-started/install) (Yarn 1 *will not work*; Yarn 2 *might work* but has not been tested)

  * [pnpm](https://pnpm.io/) may also work but has not been tested

### Installation

The instructions here are for [Yarn](https://yarnpkg.com/), but can be adapted
for [pnpm](https://pnpm.io/) or other package managers.

The following assumes you're using [Yarn
workspaces](https://yarnpkg.com/features/workspaces) from the root of your
Drupal installation, i.e. where your root `composer.json` is located, one level
above the web directory. You must also be using [the Plug'n'Play install
strategy](https://yarnpkg.com/features/pnp), which is the default in modern
Yarn. You'll need to have a `package.json` in the root directory containing a
`workspaces` array:

```json
{
  "workspaces": [
    "web/modules/contrib/refreshless/**!(vendor)/*"
  ]
}
```

If you already have other workspace paths defined, just merge that in.

Then, run `yarn install` and Turbo will be downloaded and extracted to the
correct location. This is accomplished via the
[`@consensus.enterprises/pnp-vendorize`](https://www.npmjs.com/package/@consensus.enterprises/pnp-vendorize)
package.

### Patching

The following describes how to apply our Turbo patches [via
Yarn](https://yarnpkg.com/features/patching) but can also be [adapted for
pnpm](https://pnpm.io/cli/patch). We want to eventually contribute to Turbo and
remove the need for patches, but until that time, patching will be required if
you need fixes for various issues. Note that Yarn doesn't allow including a
`"resolutions"` field in anything but the top-level `package.json`, so you have
to merge this yourself into yours:

```json
"resolutions": {
  "@hotwired/turbo@8.0.10": "patch:@hotwired/turbo@npm%3A8.0.10#./web/modules/contrib/refreshless/modules/refreshless_turbo/patches/@hotwired/turbo/turbo-8-refreshless-combined.patch"
}
```

Then run `yarn install` and Yarn will patch Turbo for you.

Note that this assumes RefreshLess is installed to `web/modules/contrib` so
update the path if needed. Also note that the `"resolutions"` field only
supports exact versions and not version ranges, so when Turbo is upgraded to a
new version - including a minor version - you'll have to update both instances
of the hard-coded version above, in both the key and the value.

### Patches

#### Turbo

* [In-page anchors and `hashchange` events can cause full page visits or otherwise behave unexpectedly](https://www.drupal.org/project/refreshless/issues/3416085); individual patch is `patches/@hotwired/turbo/turbo-8-pull-1125-anchor-hashchange.patch` which is included in our combined patch.

* [Visiting pages with 4xx response codes causes re-evaluation of `<head>` JavaScript](https://www.drupal.org/project/refreshless/issues/3422964); individual patch is `patches/@hotwired/turbo/turbo-8-issue-1190-render-error-script-eval.patch` which is included in our combined patch.

* [Firefox scrolling to in-page anchor is instant and abrupt with smooth scrolling enabled](https://github.com/hotwired/turbo/issues/1255); individual patch is `patches/@hotwired/turbo/turbo-8-issue-1255-smooth-scroll-hash.patch` which is included in our combined patch.
